abstract class Program {
    public abstract <A> A accept (ProgramVisitor<A> v);
}

class Prog extends Program {
    public List<Statement> body;
    public Prog (List<Statement> body) {
        this.body = body;
    }
    public <A> A accept (ProgramVisitor<A> v) {
        return v.visit (this);
    }
}

interface ProgramVisitor<A> {
    public A visit (Prog p);
}

abstract class Statement {
    public abstract <A> A accept (StatementVisitor<A> v);
}

class CompoundStatement extends Statement {
    public Statement stm1;
    public Statement stm2;
    public CompoundStatement (Statement stm1, Statement stm2) {
        this.stm1 = stm1;
        this.stm2 = stm2;
    }
    public <A> A accept (StatementVisitor<A> v) {
        return v.visit (this);
    }
}

class AssignStatement extends Statement {
    public String id;
    public Expression exp;
    public AssignStatement (String id, Expression exp) {
        this.id = id;
        this.exp = exp;
    }
    public <A> A accept (StatementVisitor<A> v) {
        return v.visit (this);
    }
}

class PrintStatement extends Statement {
    public List<Expression> exps;
    public PrintStatement (List<Expression> exps) {
        this.exps = exps;
    }
    public <A> A accept (StatementVisitor<A> v) {
        return v.visit (this);
    }
}

interface StatementVisitor<A> {
    public A visit (CompoundStatement s);
    public A visit (AssignStatement s);
    public A visit (PrintStatement s);
}

abstract class Expression {
    public abstract <A> A accept (ExpressionVisitor<A> v);
}

class IdExpression extends Expression {
    public String id;
    public IdExpression (String id) {
        this.id = id;
    }
    public <A> A accept (ExpressionVisitor<A> v) {
        return v.visit (this);
    }
}

class NumExpression extends Expression {
    public Integer num;
    public NumExpression (Integer num) {
        this.num = num;
    }
    public <A> A accept (ExpressionVisitor<A> v) {
        return v.visit (this);
    }
}

class OpExpression extends Expression {
    public Expression left;
    public Binop op;
    public Expression right;
    public OpExpression (Expression left, Binop op, Expression right) {
        this.left = left;
        this.op = op;
        this.right = right;
    }
    public <A> A accept (ExpressionVisitor<A> v) {
        return v.visit (this);
    }
}

class EsepExpression extends Expression {
    public List<Statement> stm;
    public Expression exp;
    public EsepExpression (List<Statement> stm, Expression exp) {
        this.stm = stm;
        this.exp = exp;
    }
    public <A> A accept (ExpressionVisitor<A> v) {
        return v.visit (this);
    }
}

interface ExpressionVisitor<A> {
    public A visit (IdExpression e);
    public A visit (NumExpression e);
    public A visit (OpExpression e);
    public A visit (EsepExpression e);
}

abstract class Binop {
    public abstract <A> A accept (BinopVisitor<A> v);
}

class PLUS extends Binop {
    public PLUS () {
    }
    public <A> A accept (BinopVisitor<A> v) {
        return v.visit (this);
    }
}

class MINUS extends Binop {
    public MINUS () {
    }
    public <A> A accept (BinopVisitor<A> v) {
        return v.visit (this);
    }
}

class TIMES extends Binop {
    public TIMES () {
    }
    public <A> A accept (BinopVisitor<A> v) {
        return v.visit (this);
    }
}

class DIVIDE extends Binop {
    public DIVIDE () {
    }
    public <A> A accept (BinopVisitor<A> v) {
        return v.visit (this);
    }
}

interface BinopVisitor<A> {
    public A visit (PLUS b);
    public A visit (MINUS b);
    public A visit (TIMES b);
    public A visit (DIVIDE b);
}

